package com.test.s3;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.function.Consumer;
import lombok.extern.slf4j.Slf4j;
import reactor.core.publisher.Flux;
import software.amazon.awssdk.auth.credentials.AwsCredentialsProvider;
import software.amazon.awssdk.auth.credentials.ProfileCredentialsProvider;
import software.amazon.awssdk.core.async.AsyncRequestBody;
import software.amazon.awssdk.core.async.AsyncResponseTransformer;
import software.amazon.awssdk.services.s3.S3AsyncClient;
import software.amazon.awssdk.services.s3.model.GetObjectResponse;
import software.amazon.awssdk.transfer.s3.S3TransferManager;
import software.amazon.awssdk.transfer.s3.model.CompletedUpload;
import software.amazon.awssdk.transfer.s3.model.Upload;
import software.amazon.awssdk.transfer.s3.model.UploadRequest;


import static com.test.s3.Utils.getTransferManager;

/**
 * Java 17
 *  https://github.com/aws/aws-sdk-java/issues/474
 */
@Slf4j
public class AWSTransferManager {
    //From local aws credentials/config:
    static final String PROFILE="profile";
    static final String BUCKET="bucket";
    static final String FILE ="flux.bin";
    static final String testString = "This is some test data that will get sent to S3 over and over and " +
            "over again and is about 128 chars long, no seriously... really";
    //File size will be (128 * numLinesToSend) Bytes.
    //Start to get failures somewhere after 1_000_000.
    static final int numLinesToSend=1_000_000;
    static final byte[] testBytes = testString.getBytes(StandardCharsets.UTF_8);

    public static void main (String[] args) throws IOException, InterruptedException {
        String logFile = System.getProperty("user.dir")+"/s3.log";
        Utils.initCrtLogging(logFile);
        //Use default Creds:
        //AwsCredentialsProvider credentialsProvider = DefaultCredentialsProvider.create();
        //or Select an Env Profile
        AwsCredentialsProvider credentialsProvider = ProfileCredentialsProvider.create(PROFILE);
        //Or use SSO
        //AwsCredentialsProvider credentialsProvider = Utils.getSsoCredentials(PROFILE);

        S3AsyncClient s3AsyncClient = Utils.getS3AsyncClient(credentialsProvider);
        S3TransferManager transferManager = getTransferManager(s3AsyncClient);

        //Create a feed we can 'stream' messages to:
        ConsumerFeed<byte[]> consumerFeed = new ConsumerFeed<>();
        Flux<ByteBuffer> flux = consumerFeed.flux
                .map(ByteBuffer::wrap)
                .doOnComplete(() -> log.info("Done"));

        UploadRequest.Builder req = UploadRequest.builder()
            .requestBody(AsyncRequestBody.fromPublisher(flux))
            .addTransferListener(new S3TransferListener(FILE))
            //Fix for Error we get without contentLength:
            //https://github.com/awslabs/aws-c-s3/pull/285/files#diff-5ee13dd0ee005b828bde6d60ff89582070983e3cb7e5d92389b629a9850c27f8
            .putObjectRequest(b -> b.bucket(BUCKET).key(FILE)); //.contentLength(maxSize));

        Upload upload = transferManager.upload(req.build());
        //async stuff - wait for it.
        consumerFeed.subscription.await();

        //Send some stuff:
        Consumer<byte[]> dataFeed = consumerFeed.getDataFeed();
        for (int i=0; i < numLinesToSend; i++) {
            dataFeed.accept(testBytes);
        }
        long completionSentTime= System.currentTimeMillis();
        consumerFeed.complete();

        CompletedUpload completedUpload = upload.completionFuture().join();
        double responseTime = (System.currentTimeMillis() - completionSentTime)/1000D;
        log.info("CompleteUploadResponse: {}\n\tCloseDuration: {}",
                completedUpload.response(), String.format("%.3f seconds", responseTime));

        // Validation:
        Thread.sleep(2000L);
        log.info("Getting File");
        Path filePath = Paths.get(System.getProperty("user.dir")+"/"+FILE);

        try {
            if (Files.exists(filePath)) Files.delete(filePath);
            GetObjectResponse response =
                    s3AsyncClient.getObject(o -> o.bucket(BUCKET).key(FILE),
            AsyncResponseTransformer.toFile(filePath))
                  .join();
            log.info("Resulting File Size: {}", Files.size(filePath));
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

}